#!/usr/bin/env python
'''flowtool - code flow manager

Usage:
    flowtool.py [-hdr] [-c <cfg>] [--nofmt] [-v...] [-n] [-H <host>] [-P <path>] [-D <dirs>] [-G <gits>] [-o <file>]... [COMMAND] [ARGUMENTS ...]
    flowtool.py [-hdr] [-c <cfg>] [--nofmt] [-v...] [--list|-l]

Options:
    -h, --help              print this help
    -c, --config <cfg>      use configuration file [default: flow.cfg]
    -r, --recurse           recurse up searching for a config
    -l, --list              list available commands
    -d, --debug             print debug information
    -o, --output <file>     write a log to <file>
    -v, --verbose           control how much is said about commands
    -n, --notreally         don't execute api commands
    --nofmt                 don't format (and color) output

Flow hooks:
    -H, --host <host>       override default and config host [default: ]
    -P, --path <path>       override default and config repo [default: .]
    -D, --dirs <dirs>       set default behaviour if dirs do not exist (valid: {0}) [default: abort]
    -G, --gits <gits>       set default behaviour if gits do not exist (valid: {1}) [default: abort]

Invoking flowtool without any arguments dumps the config.
'''
DIR_MODES = ['abort', 'create', 'warn', 'ignore']
GIT_MODES = ['abort', 'create', 'warn', 'ignore', 'init']
doc = __doc__.format(*["|".join(x) for x in (DIR_MODES, GIT_MODES)])

import os, sys
cmd_name = sys.argv[0]
cmd_args = sys.argv[1:]

import flib.env
args = flib.env.parse_global_args(doc, cmd_args)

from flib.output import configure_logger
log = configure_logger(cmd_name)
sys.path.append(os.curdir)
log.debug("cwd (added to pythonpath): %s" % os.path.abspath(os.curdir))
log.debug("Args:")
log.debug(args)

from flib import abort
invalid = args.invalidate({
    '--dirs': DIR_MODES,
    '--gits': GIT_MODES
})
if invalid:
    abort(log, invalid)
else:
    del invalid

config_found = False
if args.config.startswith('/') or not args.recurse:
    config_found = os.path.isfile(args.config)
    config = flib.env.parse_config(args.config, update=True)
else:
    curdir = os.path.abspath(os.curdir)
    olddir = None
    while curdir != olddir:
        log.debug('recursing: %s' % curdir)
        here = os.sep.join((curdir, args.config))
        if os.path.isfile(here):
            config = flib.env.parse_config(here, update=True)
            break
        olddir = curdir
        curdir = os.path.dirname(curdir)
    else:
        log.error('Error: No flowfile found here or in parent directories.')
        sys.exit(1)
    config_found = True
    args['recurse_dir'] = curdir
    sys.path.append(curdir)
    log.debug("flowfile found here (added to pythonpath): %s" % os.path.abspath(curdir))

log.debug("Config:")
log.debug(config)

import flib.configured


from importlib import import_module

if 'module' in config:
    log.debug("Importing default command set: flib.flows.default")
    import_module('flib.flows.default')
    log.debug("Flow module: %s" % config.module)
    config['import_deferred'] = []
    import_module(config.module)
    deferred = config.pop('import_deferred')
    for func, fn_args, fn_kwd in deferred:
        log.debug("deferred func: %s %s %s" % (func.__name__, fn_args, fn_kwd))
        func(*fn_args, **fn_kwd)


from flib.cmd import cmd_reg

log.debug("Commands registered: %s" % ' '.join(cmd_reg))

if args.list:
    from flib.output import list_commands
    list_commands(cmd_reg)
    sys.exit(0)

command = args['COMMAND']

if command is None:
    from pprint import pformat
    config_found = {True:'', False:'not '}[config_found]
    log.info("Parsed configuration: %s (%sfound)" % (args.config, config_found))
    log.info(pformat(dict(config)))
    sys.exit(0)

arguments = args['ARGUMENTS']

if command in cmd_reg:
    result = cmd_reg[command](*arguments)
    if result:
        log.info("Command resulted:")
        log.info(result)
else:
    log.error("Unknown command: %r" % command)
    log.info("Available commands are: %s" % ", ".join(cmd_reg))
    log.info("Hint: Use --list for more details on the available commands.")
    sys.exit(1)



from functools import wraps
from inspect import isfunction#, getargspec

from flib.env import parse_args

cmd_reg = {}

def add_cmd(func, alias=None):
    if alias is None:
        alias = func.__name__
    cmd_reg[alias] = func

def expose(name=None, docargs=False):
    def wrap(func):
        if not docargs:
            add_cmd(func, memo)
            return func
        else:
            @wraps(func)
            def wrapped(*args):
                return func(parse_args(func.__doc__, argv=args))
            add_cmd(wrapped, memo)
            return wrapped
    if isfunction(name):
        memo = name.__name__
        return wrap(name)
    else:
        memo = name
        return wrap

#from flib.env import args, parse_args, config

class FlowCommand(object):

    def __init__(self, *args):
        self.args = parse_args(args)
        #self.config = config



from docopt import docopt

from collections import namedtuple
args = namedtuple(
    "DefaultGlobalArgs",
    ['debug', 'verbose', 'nofmt', 'output', 'notreally', 'host', 'path', 'dirs', 'gits']
)(False, False, True, [], False, None, '.', 'abort', 'abort')


class SimpleArgs(dict):
    '''Make docopt args more accesible.'''

    def __getattr__(self, name):
        if name in self:
            return self[name]
        elif '--' + name in self:
            return self['--' + name]
        else:
            raise AttributeError

    def _invalid_msg(self, key, choices):
        errtmpl = "Error: given %s option %r is not one of: %s"
        return errtmpl % (key, self[key], ", ".join(choices))

    def invalidate(self, schema):
        for k, v in schema.items():
            if self[k] not in v:
                return self._invalid_msg(k, v)

def parse_global_args(docstring, argv):
    global args
    args = parse_args(docstring, argv, options_first=True)
    return args

def parse_args(docstring, argv, **kwd):
    if docstring is None:
        docstring = ''
    return SimpleArgs(docopt(docstring, argv, **kwd))


from configobj import ConfigObj

class SimpleConfig(ConfigObj):
    '''Make ConfigObj more accesible.'''

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, dict.__repr__(self))

    def __getattr__(self, name):
        if name in self:
            val = self[name]
            if isinstance(val, dict):
                return self.__class__(val)
            else:
                return val
        else:
            raise AttributeError(name)

config = SimpleConfig({'repo':'.'})

def parse_config(filename, update=False):
    global config
    if update:
        config.update(ConfigObj(filename))
        return config
    else:
        return SimpleConfig(ConfigObj(filename))



'''Pure python helpes'''
from collections import namedtuple

ShellResult = namedtuple("ShellResult", ['cmdline', 'cwd', 'stdout', 'stderr', 'exit_code'])


def repr_if(thing):
    string = str(thing)
    if " " in string:
        return repr(string)
    return string

def lst2cmd(lst):
    cmd = " ".join([repr_if(x) for x in lst]) if len(lst) is not 1 else lst[0]
    return cmd

import sys
def abort(log, msg, exit_code=1):
    log.error(msg)
    sys.exit(exit_code)



#!/usr/bin/env python

#from distutils.core import setup
from setuptools import setup
from setuptools import find_packages

import os

def relname(*args):
    return os.path.join(os.path.dirname(__file__), *args)

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(relname(fname)).read()

flowcfgs = ['simple.cfg', 'test.cfg']

setup(
    name='flowlib',
    version='0.6.6a',
    description='Build your own git flow!',
    author='Konstantin Martini',
    author_email='k@tuxcode.org',
    url='https://github.com/isnok/py-flowlib',
    scripts=['ft.py'],
    packages=find_packages(), #['flib', 'flib.flows'],
    data_files=[
        ('', ['README']),
        ('flib.flows', [relname('flib', 'flows', f) for f in flowcfgs]),
    ],
    package_data={'flib.flows': flowcfgs},
    #include_package_data=True,
    install_requires=['Fabric', 'configobj', 'docopt', 'sh'],
    keywords=['git','flow','shell','local','remote','commandline'],
    long_description=read('README'),
    platforms=['Debian/GNU Linux'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        #"License :: OSI Approved :: BSD License",
    ],
    )



import os
import socket
from flib.host import Host
from flib import lst2cmd
from flib import ok_sh
from flib import ShellResult

def sh2res(r):
    return ShellResult(lst2cmd(r.cmd), r.call_args['cwd'], r.stdout, r.stderr, r.exit_code)

class LocalHost(Host):

    def __init__(self):
        self._bash = ok_sh.bash.bake('-l', '-c')
        self._cp = ok_sh.cp.bake('-v')
        self.user = os.getlogin()
        self.name = socket.gethostname()
        self.login = '%s@%s' % (self.user, self.name)

    def _sh(self, cwd, *args):
        result = self._bash(lst2cmd(args), _cwd=cwd)
        return sh2res(result)

    def handle_command(self, *args):
        result = self._bash(lst2cmd(args))
        return sh2res(result)

    #def run(self, command):
        #'''emulate fabric.api.run'''
        #result = self._bash(command)
        #return sh2res(result)

    #def sudo(self, command):
        #'''emulate fabric.api.sudo'''
        #result = self._bash('sudo ' + command)
        #return sh2res(result)

    def _put(self, source, dest):
        result = self._cp(source, dest)
        return sh2res(result)

    def _get(self, source, dest):
        result = self._cp(source, dest)
        return sh2res(result)



import pytest
import sh

tool_name = './ft.py'

@pytest.fixture
def tool():
    return getattr(sh, tool_name)

_cache = {}
@pytest.fixture
def cached(tool):
    if not _cache:
        return _cache.setdefault(0, tool())
    else:
        return _cache[0]

@pytest.mark.parametrize('test_config', [
    'flows/test/flow.cfg',
    'flows/simple/flow.cfg',
])
def test_config(tool, cached, test_config):
    confed = tool('-c', test_config)
    assert confed
    assert confed.exit_code == 0
    assert confed.stdout != cached.stdout

def test_debug(tool, cached):
    debuged = tool('-d')
    assert debuged.exit_code == 0
    assert debuged
    assert debuged.stdout != cached.stdout

from sh import ErrorReturnCode_1

def test_wrongcommand(tool):
    with pytest.raises(ErrorReturnCode_1) as excinfo:
        tool('booOOooOOOoom!')

    assert "Hint:" in excinfo.value.stdout
    assert excinfo.value.exit_code == 1
